/**
 * Created by gleicher on 10/9/2015.
 */

/*
this is the "main" file - it gets loaded last - after all the objects are loaded
make sure that twgl is loaded first

it sets up the main function to be called on window.onload

 */

 var isValidGraphicsObject = function (object) {
    if(object.name === undefined) {
        console.log("warning: GraphicsObject missing name field");
        return false;
    }

    if(typeof object.draw !== "function" && typeof object.drawAfter !== "function") {
        console.log("warning: GraphicsObject of type " + object.name + " does not contain either a draw or drawAfter method");
        return false;
    }

    if(typeof object.center !== "function") {
        console.log("warning: GraphicsObject of type " + object.name + " does not contain a center method. ");
        return false;
    }

    if(typeof object.init !== "function") {
        console.log("warning: GraphicsObject of type " + object.name + " does not contain an init method. ");
        return false;
    }

    return true;
 }
window.onload = function() {
    "use strict";

    // set up the canvas and context
    var canvas = document.createElement("canvas");
    canvas.setAttribute("width",600);
    canvas.setAttribute("height",600);
    document.body.appendChild(canvas);

    // make a place to put the drawing controls - a div
    var controls = document.createElement("DIV");
    controls.id = "controls";
    document.body.appendChild(controls);

    // a switch between camera modes
    uiMode = document.createElement("select");
    uiMode.innerHTML += "<option>ArcBall</option>";
    uiMode.innerHTML += "<option>Drive</option>";
    uiMode.innerHTML += "<option>Fly</option>";
    uiMode.innerHTML += "</select>";
    controls.appendChild(uiMode);

    var resetButton = document.createElement("button");
    resetButton.innerHTML = "Reset View";
    resetButton.onclick = function() {
        // note - this knows about arcball (defined later) since arcball is lifted
        arcball.reset();

        drivePos = [0,.2,5];
        driveTheta = 0;
        driveXTheta = 0;

    }
    controls.appendChild(resetButton);

    // make some checkboxes - using my cheesy panels code
    checkboxes = makeCheckBoxes([ ["Run",1], ["Examine",0] ]); //


    // make some sliders - using my cheesy panels code
    sliders = makeSliders([["TimeOfDay",0,24,12]]);

    // this could be gl = canvas.getContext("webgl");
    // but twgl is more robust
    gl = twgl.getWebGLContext(canvas);

    // make a fake drawing state for the object initialization
    var drawingState = {
        gl : gl,
        proj : twgl.m4.identity(),
        view : twgl.m4.identity(),
        camera : twgl.m4.identity(),
        sunDirection : [0,1,0]
    }
    

    // information for the cameras
    lookAt = [0,0,0];
    lookFrom = [0,5,-10];
    fov = 1.1;

    arcball = new ArcBall(canvas);

    // for timing
    realtime = 0;
    lastTime = Date.now();

    // parameters for driving
    drivePos = [0,.2,5];
    driveTheta = 0;
    driveXTheta = 0;

    // cheesy keyboard handling
    keysdown = {};

    document.body.onkeydown = function(e) {
        var event = window.event ? window.event : e;
        keysdown[event.keyCode] = true;
        e.stopPropagation();
    };
    document.body.onkeyup = function(e) {
        var event = window.event ? window.event : e;
        delete keysdown[event.keyCode];
        e.stopPropagation();
    };

    var error = false;
    console.log("Setting up scene...");
    scene = new Scene(town_vshader_content, town_fshader_content);

    /* Make sure the shaders are compiled and linked and are valid */
    if(!error && (error |= scene.compileShaders()))
        console.log("Couldn't compile shaders.");
    if(!error && (error |= scene.linkShaders()))
        console.log("Couldn't link shaders.");
    console.log("Shaders compiled and linked successfully.");
    
    /* Set the default scene properties */
    if(!error)
    {
        scene.enableDiffuse();
        // scene.disableDiffuse();
        scene.enableSpecular();
        scene.enableAmbient();
        scene.disableLightAverage();
        // scene.enableLightAverage();
        scene.denyAlpha();
        scene.enableNormals();
    }

    /* Set uniform names for other properties */
    if(!error)
        console.log("Setting other uniform properties...");
    if(!error && (error |= scene.initGLLighting(
                    "lightingProperties", "lightPositions", "lightColors",
                    "lightIntensities", "lightFalloffs", "lightTypes",
                    "cameraPosition", "lightTx")))
        console.log("\tCouldn't initialize uniforms!! FAILURE");

    if(!error && scene.setViewUniforms("view", "proj", "cameraPosition"))
    {
        console.log("\tFAILURE: Couldn't set view, proj and camera position uniforms!");
        error = true;
    }

    if(!error && scene.setTextureUniforms("useTexture", "textureSample", "hardness"))
    {
        console.log("\tFAILURE: Couldn't set texture uniforms!");
        error = true;
    }

    if(!error && !scene.verifyGLUniforms()) 
        console.log("\tUniforms seem to be okay.");
    else {
        error = true;
        console.log("\tFAILURE: Please check uniform setup. Something isn't right.");
    }

    /* Set the attribute names */
    scene.setAttributeNames("vertices", "normals", "diffuse", "texCoords", "model");

    /* Add lights */
    if(!error) console.log("Adding lights...");
    if(!error && scene.addLight(
                [0.0, 0.0, 0.0],  // position
                [1.0, 1.0, 1.0, 1.0], // color
                .3, 200.0, "ambient"))
    {
        error = true;
        console.log("\tFAILURE");
    }
    if(!error && scene.addLight(
                //[0.0, 6.0, 0.0],  // position
                [0.0, 3.0, 0.0],  // position
                [1.0, 1.0, 1.0, 1.0], // color
                1.0, // intensity
                0.9, "point"))
    {
        error = true;
        console.log("\tFAILURE");
    }
    scene.setLightAnimationCallback(1, sun_light_animation);
    if(!error)
        console.log("\tAll lights added successfully.");

    /* Load and create all models here */
    var ground_model = new WaveObj(ground_content, "ground");
    ground_model.setHardness(1.0);
    ground_model.setGlobalColor("green", 255.0);
    if(!error && !scene.addObject(ground_model))
    {
        console.log("Ground successfully added to scene.");
    } else console.log("FAILURE: couldn't add ground model to scene!");

    var roads_model = new WaveObj(roads_content, "roads");
    roads_model.setHardness(1.0);
    roads_model.setGlobalColor("black", 255.0);
    if(!error && !scene.addObject(roads_model))
    {
        console.log("Ground successfully added to scene.");
    } else console.log("FAILURE: couldn't add roads model to scene!");

    var houses_model = new WaveObj(houses_content, "houses");
    houses_model.setHardness(2.0);
    houses_model.setGlobalColor("brown", 255.0);
    if(!error && !scene.addObject(houses_model))
    {
        console.log("Ground successfully added to scene.");
    } else console.log("FAILURE: couldn't add houses model to scene!");

    var concrete_model = new WaveObj(concrete_content, "concrete");
    concrete_model.setHardness(2.0);
    concrete_model.setGlobalColor("grey", 255.0);
    if(!error && !scene.addObject(concrete_model))
    {
        console.log("Ground successfully added to scene.");
    } else console.log("FAILURE: couldn't add concrete model to scene!");

    var sun_model = new WaveObj(sun_content, "sun");
    sun_model.setHardness(1.0);
    sun_model.setGlobalColor("yellow", 255.0);
    sun_model.setCenter([0, 6, 0]);
    sun_model.setAnimationCallback(sun_animation);
    if(!error && !scene.addObject(sun_model))
    {
        console.log("Ground successfully added to scene.");
    } else console.log("FAILURE: couldn't add sun model to scene!");

    var roundabout_model1 = new WaveObj(roundabout_content, "roundabout1");
    roundabout_model1.setHardness(1.0);
    roundabout_model1.setGlobalColor("blue", 255.0);
    roundabout_model1.setTranslation([-1.6, .1, 2.5]);
    roundabout_model1.setAnimationCallback(roundabout_animation);
    if(!error && !scene.addObject(roundabout_model1))
    {
        console.log("Ground successfully added to scene.");
    } else console.log("FAILURE: couldn't add roundabout model1 to scene!");

    var roundabout_model2 = new WaveObj(roundabout_content, "roundabout2");
    roundabout_model2.setHardness(1.0);
    roundabout_model2.setGlobalColor("blue", 255.0);
    roundabout_model2.setTranslation([-1.0, .1, -2.5]);
    roundabout_model2.setAnimationCallback(roundabout_animation);
    if(!error && !scene.addObject(roundabout_model2))
    {
        console.log("Ground successfully added to scene.");
    } else console.log("FAILURE: couldn't add roundabout model2 to scene!");

    var plane_model1 = new WaveObj(plane_content, "plane1");
    plane_model1.setCenter([2.53894, 4.01375, -0.15483]);
    plane_model1.setHardness(1.0);
    plane_model1.setRotation([0, Math.PI, 0]);
    plane_model1.setGlobalColor("grey", 255.0);
    plane_model1.setAnimationCallback(plane_animation);
    if(!error && !scene.addObject(plane_model1))
    {
        console.log("Ground successfully added to scene.");
    } else console.log("FAILURE: couldn't add plane model1 to scene!");

    var plane_model2 = new WaveObj(plane_content, "plane2");
    plane_model2.setCenter([2.53894, 4.01375, -0.15483]);
    plane_model2.setHardness(1.0);
    plane_model2.setGlobalColor("grey", 255.0);
    plane_model2.setAnimationCallback(plane_animation);
    if(!error && !scene.addObject(plane_model2))
    {
        console.log("Ground successfully added to scene.");
    } else console.log("FAILURE: couldn't add plane model2 to scene!");


    grobjects.push(ground_model);
    grobjects.push(roads_model);
    grobjects.push(houses_model);
    grobjects.push(concrete_model);
    grobjects.push(sun_model);
    grobjects.push(roundabout_model1);
    grobjects.push(roundabout_model2);
    grobjects.push(plane_model1);
    grobjects.push(plane_model2);
    // a selector for which object should be examined
    toExamine = document.createElement("select");
    grobjects.forEach(function(obj) {
           toExamine.innerHTML +=  "<option>" + obj.name + "</option>";
        });
    controls.appendChild(toExamine);

    draw();
};

// the actual draw function - which is the main "loop"
function draw() {
    // advance the clock appropriately (unless its stopped)
    var curTime = Date.now();
    var timechg = 0;
    if (checkboxes.Run.checked) {
        realtime += (curTime - lastTime);
        timechg = curTime - lastTime;
    }
    lastTime = curTime;

    // first, let's clear the screen
    gl.clearColor(0.3, 0.3, .8, 1.0);
    gl.enable(gl.DEPTH_TEST);
    gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

    // figure out the transforms
    var projM = twgl.m4.perspective(fov, 1, 0.1, 100);
    var cameraM = twgl.m4.lookAt(lookFrom,lookAt,[0,1,0]);
    var viewM = twgl.m4.inverse(cameraM);

    // implement the camera UI
    if (uiMode.value == "ArcBall") {
        viewM = arcball.getMatrix();
        twgl.m4.setTranslation(viewM, [0, 0, -10], viewM);
    } else if (uiMode.value == "Drive") {
        if (keysdown[65]) { driveTheta += .02; }
        if (keysdown[68]) { driveTheta -= .02; }
        if (keysdown[87]) {
            var dz = Math.cos(driveTheta);
            var dx = Math.sin(driveTheta);
            drivePos[0] -= .05*dx;
            drivePos[2] -= .05*dz;
        }
        if (keysdown[83]) {
            var dz = Math.cos(driveTheta);
            var dx = Math.sin(driveTheta);
            drivePos[0] += .05*dx;
            drivePos[2] += .05*dz;
        }

        cameraM = twgl.m4.rotationY(driveTheta);
        twgl.m4.setTranslation(cameraM, drivePos, cameraM);
        viewM = twgl.m4.inverse(cameraM);
    }else if (uiMode.value == "Fly") {

        if (keysdown[65] || keysdown[37]) { 
            driveTheta += .02; 
        }else if (keysdown[68] || keysdown[39]) { 
            driveTheta -= .02; 
        }

        if (keysdown[38]) { driveXTheta += .02; }
        if (keysdown[40]) { driveXTheta -= .02; }

        var dz = Math.cos(driveTheta);
        var dx = Math.sin(driveTheta);
        var dy = Math.sin(driveXTheta);

        if (keysdown[87]) {
            drivePos[0] -= .25*dx;
            drivePos[2] -= .25*dz;
            drivePos[1] += .25 * dy;
        }

        if (keysdown[83]) {
            drivePos[0] += .25*dx;
            drivePos[2] += .25*dz;
            drivePos[1] -= .25 * dy;
        }

        cameraM = twgl.m4.rotationX(driveXTheta);
        twgl.m4.multiply(cameraM, twgl.m4.rotationY(driveTheta), cameraM);
        twgl.m4.setTranslation(cameraM, drivePos, cameraM);
        viewM = twgl.m4.inverse(cameraM);
    }

    // get lighting information
    var tod = Number(sliders.TimeOfDay.value);
    var sunAngle = Math.PI * (tod-6)/12;
    var sunDirection = [Math.cos(sunAngle),Math.sin(sunAngle),0];

    // make a real drawing state for drawing
    var drawingState = {
        gl : gl,
        proj : projM,   // twgl.m4.identity(),
        view : viewM,   // twgl.m4.identity(),
        camera : cameraM,
        timeOfDay : tod,
        sunDirection : sunDirection,
        realtime : realtime
    }

    /* Inform our scene about changes to the environment */
    if(scene.setView(viewM, projM, lookFrom))
        console.log("FAILURE: Failed to set scene view!!");

    // initialize all of the objects that haven't yet been initialized (that way objects can be added at any point)
    grobjects.forEach(function(obj) { 
        if(!obj.__initialized) {
            if(isValidGraphicsObject(obj)){
                obj.init(drawingState);
                obj.__initialized = true;
            }
        }
    });

    // now draw all of the objects - unless we're in examine mode
    if (checkboxes.Examine.checked) {
        // get the examined object - too bad this is an array not an object
        var examined = undefined;
        grobjects.forEach(function(obj) { if (obj.name == toExamine.value) {examined=obj;}});
        var ctr = examined.center(drawingState);
        var shift = twgl.m4.translation([-ctr[0],-ctr[1],-ctr[2]]);
        twgl.m4.multiply(shift,drawingState.view,drawingState.view);
        
        scene.draw(true);
        examined.animate([timechg, 0]);
        examined.draw();

        // if(examined.draw) examined.draw(drawingState);
        // if(examined.drawAfter) examined.drawAfter(drawingState);
    } else {

        // grobjects.forEach(function (obj) {
            // if(obj.draw) obj.draw(drawingState);
        // });

        // grobjects.forEach(function (obj) {
            // if(obj.drawAfter) obj.drawAfter();
        // });
        scene.animate([timechg, tod]);
        scene.draw(false);
    }

    window.requestAnimationFrame(draw);
};
